-- $Id: bgs.adb,v 1.3 2006/11/27 00:14:21 baker Exp baker $

with Ada.Text_IO; use Ada.Text_IO;
with Ada.Numerics.Elementary_Functions;
with Cycles; use Cycles;
with Random_Tools; use Random_Tools;
with Sizes; use Sizes;

package body BGS is

use Event_Queue;
use Periodic_Task_Queue;
use Aperiodic_Task_Queue;
use Event_IO;
use Int_IO;
use Event_IO;

procedure Background_Server
   (Mean_Interarrival_Time: in Float;
    Aperiodic_Load        : in Float;
    Periodic_Load_List    : in Process_List;
    Avg_Resp_Time         : out Float;
    Resp_Time_Std_Dev     : out Float;
    Max_Resp_Time         : out Float;
    Requested_Task        : out integer;
    Completed_Task        : out integer;
    Missed_Deadline_Task  : out integer;
    Server_Util           : out Float) is

   Overloaded: Boolean := false;

   Mean_Exec_Time: constant Float := Aperiodic_Load * Mean_Interarrival_Time;

   E, Completion_Event  : Event;
   A_T_Info             : Aperiodic_Task_Info;
   Last_Dispatching_Time: integer := 0;
   Current_Task         : Task_ID_Number := Nul_Task;
   ET, IAT              : integer;
   Total_ExecTime       : integer := 0;

   CPU_Is_Idle          : Boolean := true;
   Last_Idle_Time       : Integer := 0;
   Next_Major_Cycle     : Integer := 0;
   Major_Cycle_Idle     : Integer := 0;

   procedure Compute_Statistics is
      use Ada.Numerics.Elementary_Functions;
   begin
      if (Aperiodic_Task_Completed > 0) then
        Resp_Time_Std_Dev := Sqrt ((Total_Sq_Resp_Time -
         (Total_Response_Time**2)/float (Aperiodic_Task_Completed))/
         float (Aperiodic_Task_Completed-1));
      else
        Resp_Time_Std_Dev := float (Erroneous);
      end if;
      Max_Resp_Time := float (Maximum_Response_Time);
      Requested_Task := Aperiodic_Task_Requested;
      Completed_Task := Aperiodic_Task_Completed;
      Missed_Deadline_Task := Aperiodic_Task_Missed;
      Server_Util := float (Total_ExecTime)/float (End_of_Test);
   end Compute_Statistics;


   procedure Compute_Response_Time is
   begin
      Response_Time := V_Time - A_T_Info.Release_Time;
      Total_Response_Time := Total_Response_Time + float (Response_Time);
      Total_Sq_Resp_Time := Total_Sq_Resp_Time + float (Response_Time)**2;
      if Response_Time > Maximum_Response_Time
      then Maximum_Response_Time := Response_Time; end if;
      Aperiodic_Task_Completed := Aperiodic_Task_Completed + 1;
      if V_time > A_T_Info.Release_Time + Server_Period
      then Aperiodic_Task_Missed := Aperiodic_Task_Missed + 1; end if;
   end Compute_Response_Time;


   procedure Update_A_T_Info is
   begin
      A_T_Info := Front_Of (A_T_Queue); Pop (A_T_Queue);
      A_T_Info.Remaining_Exec_Time := A_T_Info.Remaining_Exec_Time-Elapsed_Time;
      pragma assert (A_T_Info.Remaining_Exec_Time > 0, "updae_a_t_info");
      Add (A_T_Info, A_T_Queue);
   end Update_A_T_Info;


   procedure Try_Dispatching is
      Top: Task_ID_Number := Nul_Task;
   begin
      if not Is_Empty (P_T_Queue) then Top := Front_Of (P_T_Queue).T_ID;
      elsif not Is_Empty (A_T_Queue) then Top := Front_Of (A_T_Queue).T_ID;
      end if;

      if CPU_Is_Idle and then Top/=Nul_Task then
         Major_Cycle_Idle := Major_Cycle_Idle+ (V_Time-Last_Idle_Time);
         CPU_Is_Idle := false;
      elsif not CPU_Is_Idle and then Top=Nul_Task then
         Last_Idle_Time := V_Time;
         CPU_Is_Idle := true;
      end if;

      if Current_Task = Top then return; end if;
      if Current_Task /= Nul_Task then
         Delete (Completion_Event, E_Queue);
         Elapsed_Time := V_Time - Last_Dispatching_Time;
         if Current_Task in Periodic_Task_ID_Number then
            P_T_List (Current_Task).Remaining_Exec_Time :=
            P_T_List (Current_Task).Remaining_Exec_Time - Elapsed_Time;
         else Update_A_T_Info; end if;
      end if;

      Current_Task := Top;
      Completion_Event.E_ID := Current_Task;
      if Current_Task in Periodic_Task_ID_Number then
         Completion_Event.Kind := Periodic_Completion;
         Completion_Event.E_Time :=
         V_Time + P_T_List (Current_Task).Remaining_Exec_Time;
      else
         Completion_Event.Kind := Aperiodic_Completion;
         Completion_Event.E_Time :=
         V_Time + Front_Of (A_T_Queue).Remaining_Exec_Time;
      end if;
      Add (Completion_Event, E_Queue);
      Last_Dispatching_Time := V_Time;
   end Try_Dispatching;

begin -- Background_Server --

--------------------
-- Initialization --
--------------------

Clear (E_Queue); Clear (P_T_Queue); Clear (A_T_Queue);
for I in Periodic_Task_ID_Number loop
   P_T_List (I) := (0, Periodic_Load_List (I).Exec_Time,
      Periodic_Load_List (I).Period, Periodic_Load_List (I).Exec_Time);
end loop;
for I in Periodic_Task_ID_Number loop
   Add ((I, Periodic_Request, 0), E_Queue);
end loop;
Add ((Aperiodic_Task_ID_Number'first, Aperiodic_Request,
     Exp_Dist (Mean_Interarrival_Time)), E_Queue);

--------------------------
-- Main simulation loop --
--------------------------

   Init_Times;

loop
   while Front_Of (E_Queue).E_Time = V_Time loop
      E := Front_Of (E_Queue); Pop (E_Queue);
      case E.Kind is
         when Periodic_Request =>
            -----------------------------------------------------------
            -- insert the periodic task into the periodic task queue --
            -- insert deadline event of this task into event queue   --
            -----------------------------------------------------------
            Add ((E.E_ID, -P_T_List (E.E_ID).Deadline), P_T_Queue);
            Add ((E.E_ID, Periodic_Deadline, P_T_List (E.E_ID).Deadline), E_Queue);

         when Aperiodic_Request =>
            ----------------------------------------------------------------
            -- insert the aperiodic request into the aperiodic task queue --
            -- insert next aperiodic request event into the event queue   --
            ----------------------------------------------------------------
            Aperiodic_Task_Requested := Aperiodic_Task_Requested + 1;
            ET := Exp_Dist (Mean_Exec_Time);
            if ET=0 then
              Aperiodic_Task_Completed := Aperiodic_Task_Completed+1;
            else
              Add ((E.E_ID, V_Time, ET, ET), A_T_Queue);
              Total_ExecTime := Total_ExecTime+ET;
            end if;
            IAT := Exp_Dist (Mean_Interarrival_Time);
            Add ((E.E_ID+1, Aperiodic_Request, V_Time+IAT), E_Queue);

         when Periodic_Completion =>
            ---------------------------------------------------------
            -- remove deadline event of this task from event queue --
            -- remove task from periodic task queue                --
            -- update periodic task list
            -- generate next periodic request event --
            -- and insert it into the event queue   --
            -- set processor to idle --
            ---------------------------------------------------------
            Delete ((E.E_ID, Periodic_Deadline, P_T_List (E.E_ID).Deadline), E_Queue);
            Pop (P_T_Queue);
            P_T_List (E.E_ID).Release_Time := P_T_List (E.E_ID).Release_Time +
               Periodic_Load_List (E.E_ID).Period;
            P_T_List (E.E_ID).Deadline := P_T_List (E.E_ID).Deadline +
               Periodic_Load_List (E.E_ID).Period;
            P_T_List (E.E_ID).Remaining_Exec_Time :=
               Periodic_Load_List (E.E_ID).Exec_Time;
            Add ((E.E_ID, Periodic_Request, P_T_List (E.E_ID).Release_Time), E_Queue);
            Current_Task := Nul_Task;

         when Aperiodic_Completion =>
            -----------------------------
            -- calculate response time --
            -- set processor to idle --
            -----------------------------
            A_T_Info := Front_Of (A_T_Queue); Pop (A_T_Queue);
            Compute_Response_Time;
            Current_Task := Nul_Task;

         when Periodic_Deadline => raise Schedule_Fail;

         when Aperiodic_Server_Replenishment => null;
      end case;

      if V_Time >= End_of_Test
         and then Front_Of (E_Queue).Kind = Periodic_Request
         then raise Schedule_Complete; end if;
      if V_Time>=Next_Major_Cycle then
         if CPU_Is_Idle then
            Major_Cycle_Idle := Major_Cycle_Idle+ (V_Time-Last_Idle_Time);
            Last_Idle_Time := V_Time;
         end if;
         if Major_Cycle_Idle=0 and then V_Time/=0 and then
            not Overloaded then
            put (Errs, "BGS: "); Overloaded := true;
            Int_IO.put (Errs, V_Time, 8);
            Flt_IO.put (Errs, Mean_Interarrival_Time, 4, 2, 0);
            Flt_IO.put (Errs, Aperiodic_Load, 4, 2, 0);
            put_line (Errs, " CPU overload ***");
         end if;
         Major_Cycle_Idle := 0;
         Next_Major_Cycle := Next_Major_Cycle+Major_Period;
      end if;
   end loop;
   Try_Dispatching;
   if Front_Of (E_Queue).E_Time < V_Time then raise Program_Error; end if;
   V_Time := Front_Of (E_Queue).E_Time;
end loop;

exception
   when Schedule_Complete =>
      while not Is_Empty (A_T_Queue) and then
      Front_Of (A_T_Queue).Release_Time < V_Time - Server_Period loop
         Aperiodic_Task_Missed := Aperiodic_Task_Missed +1; Pop (A_T_Queue);
      end loop;
      -- compute average response time
      if Aperiodic_Task_Completed > 0
      then Avg_Resp_Time := Total_Response_Time/float (Aperiodic_Task_Completed);
      else Avg_Resp_Time := float (Erroneous);
        Aperiodic_Task_Completed := Erroneous;
        Total_Sq_Resp_Time := 10.0E22;
      end if;
      Compute_Statistics;

   when Schedule_Fail =>
      while not Is_Empty (A_T_Queue) and then
      Front_Of (A_T_Queue).Release_Time < V_Time - Server_Period loop
         Aperiodic_Task_Missed := Aperiodic_Task_Missed +1; Pop (A_T_Queue);
      end loop;
      -- display message
      put_line (Errs,
        "BGS: Periodic task # " & Periodic_Task_ID_Number'image (E.E_ID)
        & " misses its deadline at time " & integer'image (V_Time));
      Avg_Resp_Time := float (Erroneous);
      Compute_Statistics;
   when others =>
      put (Errs, "BGS: time: ");put (Errs, Integer'image (V_Time));
      put (Errs, " e_time: ");
      put (Errs, Front_Of (E_Queue).E_Time, 7); put (Errs, " (");
      put (Errs, " (");
      put (Errs, Front_Of (E_Queue).Kind);
      put (Errs, ")"); new_line (Errs);
end Background_Server;

end BGS;



