-- $Id: dds.adb,v 1.3 2006/11/27 00:14:21 baker Exp baker $

-- fixed-priority deferrable server ("rms" version, no background)
-- Baker re-write, based on deadline version

with Ada.Text_IO; use Ada.Text_IO;
with Ada.Float_Text_IO; use Ada.Float_Text_IO;
with Ada.Numerics.Elementary_Functions;
with Cycles; use Cycles;
with Random_Tools; use Random_Tools;
with Sizes; use Sizes;

package body DDS is

use Event_Queue;
use Periodic_Task_Queue;
use Aperiodic_Task_Queue;
use Replenishment_Queue;
use Event_IO;
use Int_IO;

procedure DD_Server
   (Mean_Interarrival_Time: in Float;
    Aperiodic_Load        : in Float;
    Periodic_Load_List    : in Process_List;
    Server_Exec_Time      : in integer;
    Avg_Resp_Time         : out Float;
    Resp_Time_Std_Dev     : out Float;
    Max_Resp_Time         : out Float;
    Requested_Task        : out integer;
    Completed_Task        : out integer;
    Missed_Deadline_Task  : out integer;
    Server_Util           : out Float) is

   Overloaded: Boolean := false;

   Mean_Exec_Time: constant Float := Aperiodic_Load*Mean_Interarrival_Time;
   Server_ID     : constant Task_ID_Number := Task_ID_Number'first;

   Server_Priority      : integer := -Server_Period;
   E, Completion_Event  : Event;
   A_T_Info             : Aperiodic_Task_Info;
   R_Info               : Replenishment_Info;
   Elapsed_Time,
   Last_Dispatching_Time: integer := 0;
   Current_Task         : Task_ID_Number := Nul_Task;
   ET, IAT              : Integer;
   Server_Eligible,
   Undefined            : constant Integer := Integer'first;
   t_z                  : integer := 0;

   CPU_Is_Idle          : Boolean := true;
   Last_Idle_Time       : Integer := 0;
   Next_Major_Cycle     : Integer := 0;
   Major_Cycle_Idle     : Integer := 0;

   procedure Compute_Statistics is
      use Ada.Numerics.Elementary_Functions;
   begin
      if (Aperiodic_Task_Completed > 0) then
        Resp_Time_Std_Dev := Sqrt ((Total_Sq_Resp_Time -
         (Total_Response_Time**2)/float (Aperiodic_Task_Completed))/
         float (Aperiodic_Task_Completed-1));
      else
        Resp_Time_Std_Dev := float (Erroneous);
      end if;
      Max_Resp_Time := float (Maximum_Response_Time);
      Requested_Task := Aperiodic_Task_Requested;
      Completed_Task := Aperiodic_Task_Completed;
      Missed_Deadline_Task := Aperiodic_Task_Missed;
      Server_Util := float (Total_Server)*100.00/float (V_Time);
   end Compute_Statistics;


   procedure Compute_Response_Time is
   begin
      Response_Time := V_Time - A_T_Info.Release_Time;
      Total_Response_Time := Total_Response_Time+float (Response_Time);
      Total_Sq_Resp_Time := Total_Sq_Resp_Time+float (Response_Time)**2;
      if Response_Time > Maximum_Response_Time
      then Maximum_Response_Time := Response_Time; end if;
      Aperiodic_Task_Completed := Aperiodic_Task_Completed + 1;
      if V_time > A_T_Info.Release_Time + Server_Period
      then Aperiodic_Task_Missed := Aperiodic_Task_Missed + 1; end if;
   end Compute_Response_Time;

-- Notes:

-- The server is ELIGIBLE (to execute in the foreground) if all
-- of the following are true:
--  i.   There is an aperiodic request pending.
--  ii.  There is a chunk of server time with R_Time<=V_Time.
--  iii. There is a periodic job competing for the processor.
-- Condition (iii) applies only if we are allowing the server
-- to use background time, which we are not, in this version.

   procedure Try_Dispatching is
      Top: Task_ID_Number := Nul_Task;
      P: integer;
   begin

      if Is_Empty (A_T_Queue) then
         if Is_Empty (P_T_Queue) then

            -- (1) idle task
            Top := Nul_Task;
         else

            -- (2) periodic task, no aperiodic competition
            Top := Front_Of (P_T_Queue).T_ID;
         end if;
      elsif Is_Empty (P_T_Queue) then
         if R_Info.R_Amount= 0 then

            -- (3a) idle task
            Top := Nul_Task;
         else

            -- (3b) aperiodic task
            Top := Front_Of (A_T_Queue).T_ID;
         end if;
      elsif R_Info.R_Amount=0 then

         -- (4) periodic task, could preempt background server
         Top := Front_Of (P_T_Queue).T_ID;
      else -- server is eligible
         P := Front_Of (P_T_Queue).Priority;
         if P<=Server_Priority then

            -- (5) server in foreground, preempting periodic task
            Top := Front_Of (A_T_Queue).T_ID;
         else

            -- (6) periodic task, preempting foreground server
            Top := Front_Of (P_T_Queue).T_ID;
         end if;
      end if;

      if CPU_Is_Idle and then Top/=Nul_Task then
         Major_Cycle_Idle := Major_Cycle_Idle+ (V_Time-Last_Idle_Time);
         CPU_Is_Idle := false;
      elsif not CPU_Is_Idle and then Top=Nul_Task then
         Last_Idle_Time := V_Time;
         CPU_Is_Idle := true;
      end if;

      if Top=Current_Task
      then return;
      end if;

      Delete (Completion_Event, E_Queue);

      if Current_Task/=Nul_Task then
         Elapsed_Time := V_Time - Last_Dispatching_Time;
         pragma Assert (Elapsed_Time>0, "dispatch (2)");
         if Current_Task in Periodic_Task_ID_Number then
            P_T_List (Current_Task).Remaining_Exec_Time :=
            P_T_List (Current_Task).Remaining_Exec_Time - Elapsed_Time;
         else -- Aperiodic Task
            R_Info.R_Amount := R_Info.R_Amount - Elapsed_Time;
            Total_Server := Total_Server + Elapsed_Time;
            A_T_Info := Front_Of (A_T_Queue); Pop (A_T_Queue);
            A_T_Info.Remaining_Exec_Time :=
              A_T_Info.Remaining_Exec_Time-Elapsed_Time;
            pragma Assert (A_T_Info.Remaining_Exec_Time>0, "dispatch (3)");
            Add (A_T_Info, A_T_Queue);
         end if;
      end if;

      Current_Task := Top;
      if Current_Task=Nul_Task then return; end if;

      Completion_Event.E_ID := Current_Task;
      if Current_Task in Periodic_Task_ID_Number then
         Completion_Event.Kind := Periodic_Completion;
         Completion_Event.E_Time :=
         V_Time + P_T_List (Current_Task).Remaining_Exec_Time;
      else
         Completion_Event.Kind := Aperiodic_Completion;
         A_T_Info := Front_Of (A_T_Queue);
         if A_T_Info.Remaining_Exec_Time > R_Info.R_Amount then
            Completion_Event.E_Time := V_Time + R_Info.R_Amount;
         else Completion_Event.E_Time := V_Time+A_T_Info.Remaining_Exec_Time;
         end if;
      end if;
      Last_Dispatching_Time := V_Time;
      Add (Completion_Event, E_Queue);

   end Try_Dispatching;

begin

   Clear (E_Queue); Clear (P_T_Queue); Clear (A_T_Queue); Clear (R_Queue);

   if Aperiodic_Load>=float (Server_Exec_Time)/float (Server_Period) then
      Total_Sq_Resp_Time := 0.0;
      Total_Response_Time := 0.0;
      Aperiodic_Task_Completed := Erroneous;
      Aperiodic_Task_Requested :=0;
      Aperiodic_Task_Missed := 0;
      Total_Server := 0;
      V_Time := 1;
      raise Schedule_Complete;
   end if;

   for I in Periodic_Task_ID_Number loop
      P_T_List (I) := (-Periodic_Load_List (I).Period,
                  Periodic_Load_List (I).Exec_Time, 0, 0);
   end loop;
   for I in Periodic_Task_ID_Number loop
      Add ((I, Periodic_Request, 0), E_Queue);
   end loop;
   Add ((Aperiodic_Task_ID_Number'first, Aperiodic_Request,
        Exp_Dist (Mean_Interarrival_Time)), E_Queue);
   Add ((Server_ID, Aperiodic_Server_Replenishment, 0), E_Queue);
   R_Info.R_Amount := 0;

   Init_Times;

loop

   while Front_Of (E_Queue).E_Time = V_Time loop
      E := Front_Of (E_Queue); Pop (E_Queue);
      case E.Kind is
         when Periodic_Request =>
            P_T_List (E.E_ID).Release_Time := P_T_List (E.E_ID).Release_Time +
               Periodic_Load_List (E.E_ID).Period;
            P_T_List (E.E_ID).Deadline := P_T_List (E.E_ID).Deadline +
               Periodic_Load_List (E.E_ID).Period;
            P_T_List (E.E_ID).Remaining_Exec_Time :=
               Periodic_Load_List (E.E_ID).Exec_Time;
            Add ((E.E_ID, -Periodic_Load_List (E.E_ID).Period), P_T_Queue);
            Add ((E.E_ID, Periodic_Deadline, P_T_List (E.E_ID).Deadline), E_Queue);
            Add ((E.E_ID, Periodic_Request, P_T_List (E.E_ID).Release_Time +
                 Periodic_Load_List (E.E_ID).Period), E_Queue);

         when Aperiodic_Request =>
            Aperiodic_Task_Requested := Aperiodic_Task_Requested+1;
            ET := Exp_Dist (Mean_Exec_Time);
            if ET=0 then
               Aperiodic_Task_Completed :=Aperiodic_Task_Completed+1;
            else
               Add ((E.E_ID, V_Time, ET, ET), A_T_Queue);
            end if;
            IAT := Exp_Dist (Mean_Interarrival_Time);
            Add ((E.E_ID+1, Aperiodic_Request, V_Time+IAT), E_Queue);

         when Periodic_Completion =>
            Delete ((E.E_ID, Periodic_Deadline, P_T_List (E.E_ID).Deadline),
                   E_Queue);
            Pop (P_T_Queue);
            Current_Task := Nul_Task;

         when Aperiodic_Completion =>
            Elapsed_Time := V_Time - Last_Dispatching_Time;
            R_Info.R_Amount := R_Info.R_Amount - Elapsed_Time;
            Total_Server := Total_Server + Elapsed_Time;
            A_T_Info := Front_Of (A_T_Queue); Pop (A_T_Queue);
            A_T_Info.Remaining_Exec_Time :=
               A_T_Info.Remaining_Exec_Time-Elapsed_Time;
            if A_T_Info.Remaining_Exec_Time>0 then Add (A_T_Info, A_T_Queue);
            else
               Compute_Response_Time;
            end if;
            Current_Task := Nul_Task;

         when Periodic_Deadline => raise Schedule_Fail;

         when Aperiodic_Server_Replenishment =>
            R_Info.R_Amount := Server_Exec_Time;
            Add ((Server_ID, Aperiodic_Server_Replenishment,
                V_Time+Server_Period), E_Queue);

      end case;
      if V_Time>=End_of_Test and then Front_Of (E_Queue).Kind=Periodic_Request
      then raise Schedule_Complete; end if;
      if V_Time>=Next_Major_Cycle then
         if CPU_Is_Idle then
            Major_Cycle_Idle := Major_Cycle_Idle+ (V_Time-Last_Idle_Time);
            Last_Idle_Time := V_Time;
         end if;
         if Major_Cycle_Idle=0 and then V_Time>0 and then
            not Overloaded then
            put (Errs, "DDS: "); Overloaded := true;
            Int_IO.put (Errs, V_Time, 8);
            Flt_IO.put (Errs, Mean_Interarrival_Time, 4, 2, 0);
            Flt_IO.put (Errs, Aperiodic_Load, 4, 2, 0);
            put_line (Errs, " CPU overload ***");
         end if;
         Major_Cycle_Idle := 0;
         Next_Major_Cycle := Next_Major_Cycle+Major_Period;
      end if;
   end loop;

   Try_Dispatching;
   if Front_Of (E_Queue).E_Time < V_Time then raise Program_Error; end if;
   V_Time := Front_Of (E_Queue).E_Time;
end loop;

exception
   when Schedule_Complete =>
      while (not Is_Empty (A_T_Queue)) and then
      Front_Of (A_T_Queue).Release_Time < V_Time-Server_Period loop
         Aperiodic_Task_Missed := Aperiodic_Task_Missed + 1; Pop (A_T_Queue);
      end loop;
      -- compute average response time --
      if Aperiodic_Task_Completed > 0
      then Avg_Resp_Time := Total_Response_Time/float (Aperiodic_Task_Completed);
      else Avg_Resp_Time := float (Erroneous); end if;
      Compute_Statistics;

   when Schedule_Fail =>
      while (not Is_Empty (A_T_Queue)) and then
      Front_Of (A_T_Queue).Release_Time < V_Time - Server_Period loop
         Aperiodic_Task_Missed := Aperiodic_Task_Missed +1; Pop (A_T_Queue);
      end loop;
      -- display message --
      put_line (Errs,
        "DDS: Periodic task #" & Periodic_Task_ID_Number'image (E.E_ID)
        & " misses its deadline at time" & integer'image (V_Time));
      put (Errs, "remaining exec time ");
      Int_IO.put (Errs, P_T_List(E.E_ID).Remaining_Exec_Time, 8);
      new_line (Errs);
      Avg_Resp_Time := float (Erroneous);
      Compute_Statistics;

   when program_error =>
      put (Errs, "DDS: time: ");put (Errs, V_Time, 7);put (Errs, " e_time: ");
      put (Errs, Front_Of (E_Queue).E_Time, 7);put (Errs, " (");
      put (Errs, Front_Of (E_Queue).Kind);put (Errs, ")"); new_line (Errs);
end DD_Server;

end DDS;
