-- $Id: pls.adb,v 1.3 2006/11/27 00:14:21 baker Exp baker $

-- fixed-priority polling server ("rms" version, no background)
-- Baker re-write, based on deadline polling server

with Ada.Text_IO; use Ada.Text_IO;
with Ada.Float_Text_IO; use Ada.Float_Text_IO;
with Ada.Numerics.Elementary_Functions;
with Ada.Exceptions; use Ada.Exceptions;
with Cycles; use Cycles;
with Random_Tools; use Random_Tools;
with Sizes; use Sizes;

package body PLS is

use Event_Queue;
use Periodic_Task_Queue;
use Aperiodic_Task_Queue;
use Replenishment_Queue;
use Event_IO;
use Int_IO;

procedure Polling_Server
   (Mean_Interarrival_Time: in Float;
    Aperiodic_Load        : in Float;
    Periodic_Load_List    : in Process_List;
    Server_Exec_Time      : in integer;
    Avg_Resp_Time         : out Float;
    Resp_Time_Std_Dev     : out Float;
    Max_Resp_Time         : out Float;
    Requested_Task        : out integer;
    Completed_Task        : out integer;
    Missed_Deadline_Task  : out integer;
    Server_Util           : out Float) is

   Overloaded: Boolean := false;

   Mean_Exec_Time: constant Float := Aperiodic_Load*Mean_Interarrival_Time;
   Server_ID     : constant Task_ID_Number := Task_ID_Number'first;

   Server_Priority      : integer := -Server_Period;
   E, Completion_Event  : Event;
   A_T_Info             : Aperiodic_Task_Info;
   R_Info               : Replenishment_Info;
   ET, IAT              : Integer;
   Server_Eligible,
   Undefined            : constant Integer := Integer'first;
   t_z                  : integer := 0;

   CPU_Is_Idle          : Boolean := true;
   Last_Idle_Time       : Integer := 0;
   Next_Major_Cycle     : Integer := 0;
   Major_Cycle_Idle     : Integer := 0;

   procedure Compute_Statistics is
      use Ada.Numerics.Elementary_Functions;
   begin
      if (Aperiodic_Task_Completed > 0) then
        Resp_Time_Std_Dev := Sqrt ((Total_Sq_Resp_Time -
         (Total_Response_Time**2)/float (Aperiodic_Task_Completed))/
         float (Aperiodic_Task_Completed-1));
      else
        Resp_Time_Std_Dev := float (Erroneous);
      end if;
      Max_Resp_Time := float (Maximum_Response_Time);
      Requested_Task := Aperiodic_Task_Requested;
      Completed_Task := Aperiodic_Task_Completed;
      Missed_Deadline_Task := Aperiodic_Task_Missed;
      Server_Util := float (Total_Server) * 100.00 / float (V_Time);
   end Compute_Statistics;

   procedure Compute_Response_Time is
   begin
      Response_Time := V_Time - A_T_Info.Release_Time;
      Total_Response_Time := Total_Response_Time+float (Response_Time);
      Total_Sq_Resp_Time := Total_Sq_Resp_Time+float (Response_Time)**2;
      if Response_Time > Maximum_Response_Time
      then Maximum_Response_Time := Response_Time; end if;
      Aperiodic_Task_Completed := Aperiodic_Task_Completed + 1;
      if V_time > A_T_Info.Release_Time + Server_Period
      then Aperiodic_Task_Missed := Aperiodic_Task_Missed + 1; end if;
   end Compute_Response_Time;

-- Notes:

-- The server is ELIGIBLE (to execute in the foreground) if all
-- of the following are true:
--  i.   There is an aperiodic request pending.
--  ii.  There is a chunk of server time with R_Time<=V_Time.
--  iii. There is a periodic job competing for the processor.
-- Condition (iii) applies only if we are allowing the server
-- to use background time, which we don't allow, in this version.

   procedure Try_Dispatching is
      Top: Task_ID_Number := Nul_Task;
      P: integer;
   begin

      if Is_Empty (A_T_Queue) then
         if Is_Empty (P_T_Queue) then

            -- (1) idle task
            Top := Nul_Task;
         else

            -- (2) periodic task, no aperiodic competition
            Top := Front_Of (P_T_Queue).T_ID;
         end if;
         R_Info.R_Amount := 0;
      elsif Is_Empty (P_T_Queue) then
         if R_Info.R_Amount = 0 then

            -- (3a) idle task
            Top := Nul_Task;
         else

            -- (3b) aperiodic task
            Top := Front_Of (A_T_Queue).T_ID;
         end if;
      elsif R_Info.R_Amount = 0 then

         -- (4) periodic task
         Top := Front_Of (P_T_Queue).T_ID;
      else -- server is eligible
         P := Front_Of (P_T_Queue).Priority;
         if P<=Server_Priority then

            -- (5) server in foreground, preempting periodic task
            Top := Front_Of (A_T_Queue).T_ID;
         else

            -- (6) periodic task, preempting foreground server
            Top := Front_Of (P_T_Queue).T_ID;
         end if;
      end if;

      if CPU_Is_Idle and then Top /= Nul_Task then
         -- transition from idle to non-idle system
         Total_Idle := Total_Idle + (V_Time - Last_Idle_Time);
         Check_Times ("ck1");
         CPU_Is_Idle := false;
      elsif not CPU_Is_Idle and then Top = Nul_Task then
         -- transition from non-idle to idle system
         Last_Idle_Time := V_Time;
         CPU_Is_Idle := true;
      end if;

      if Top = Current_Task then
         --  no need to preempt the current task
         return;
      end if;

      --  preempt the current task

      --  the task won't be able to complete at the time expected
      Delete (Completion_Event, E_Queue);

      if Current_Task /= Nul_Task then
         --  compute how long the current task has been running
         Elapsed_Time := V_Time - Last_Dispatching_Time;
         if Current_Task in Periodic_Task_ID_Number then
            --  we are preempting a periodic task
            --  reduce its remaining execution time
            --  and increase its accumulated execution time
            --  by the amount it has just completed
            P_T_List (Current_Task).Remaining_Exec_Time :=
              P_T_List (Current_Task).Remaining_Exec_Time - Elapsed_Time;
            Total_Periodic := Total_Periodic + Elapsed_Time;
            Check_Before_Task_Switch; -- ASSERTION-CHECKING
         else -- we are preeempting an aperiodic task
            --  charge the server for the time it has just used
            R_Info.R_Amount := R_Info.R_Amount - Elapsed_Time;
            Total_Server := Total_Server + Elapsed_Time;
            Check_Times ("ck3");  -- ASSERTION-CHECKING
            --  decrement the remaining execution time
            --  of this aperiodic job
            A_T_Info := Front_Of (A_T_Queue); Pop (A_T_Queue);
            A_T_Info.Remaining_Exec_Time :=
              A_T_Info.Remaining_Exec_Time - Elapsed_Time;
            Add (A_T_Info, A_T_Queue);
         end if;
      end if;

      --  the "context switch"
      Current_Task := Top;

      --  no need to do anything more, if this is the idle task
      if Current_Task=Nul_Task then return; end if;

      --  otherwise, set up the tentative completion event
      --  for this job

      Completion_Event.E_ID := Current_Task;
      if Current_Task in Periodic_Task_ID_Number then
         Completion_Event.Kind := Periodic_Completion;
         Completion_Event.E_Time :=
           V_Time + P_T_List (Current_Task).Remaining_Exec_Time;
      else
         Completion_Event.Kind := Aperiodic_Completion;
         A_T_Info := Front_Of (A_T_Queue);
         if A_T_Info.Remaining_Exec_Time > R_Info.R_Amount then
            Completion_Event.E_Time := V_Time + R_Info.R_Amount;
         else Completion_Event.E_Time := V_Time+A_T_Info.Remaining_Exec_Time;
         end if;
      end if;
      Add (Completion_Event, E_Queue);

      Last_Dispatching_Time := V_Time;

   end Try_Dispatching;

begin

   Clear (E_Queue); Clear (P_T_Queue); Clear (A_T_Queue); Clear (R_Queue);

   if Aperiodic_Load>=float (Server_Exec_Time)/float (Server_Period) then
      Total_Sq_Resp_Time := 0.0;
      Total_Response_Time := 0.0;
      Aperiodic_Task_Completed := Erroneous;
      Aperiodic_Task_Requested :=0;
      Aperiodic_Task_Missed := 0;
      Total_Server := 0;
      V_Time := 1;
      raise Schedule_Complete;
   end if;

   --  Set up parameters for the periodic tasks, which
   --  are all assumed to be released together at time zero.
   for I in Periodic_Task_ID_Number loop
      --  Pretend that they were last releasted at time -Period,
      --  and all completed last with deadline zero,
      --  so they will have jobs arriving at time zero.
      P_T_List (I) := (Release_Time => -Periodic_Load_List (I).Period,
                       Exec_Time => Periodic_Load_List (I).Exec_Time,
                       Deadline => 0,
                       Remaining_Exec_Time =>0);
   end loop;

   --  Set up events for the next arrival of each periodic task.
   for I in Periodic_Task_ID_Number loop
      Add ((I, Periodic_Request, 0), E_Queue);
   end loop;

   --  Set up an arrival event for the aperiodic task
   Add ((Aperiodic_Task_ID_Number'first, Aperiodic_Request,
         Exp_Dist (Mean_Interarrival_Time)), E_Queue);

   --  Schedule a replenishment for the aperiodic server,
   --  to occur at time zero.
   Add ((Server_ID, Aperiodic_Server_Replenishment, 0), E_Queue);
   R_Info.R_Amount := 0;

   Init_Times;

   loop

      --  process events that occur at the same virtual time
      while Front_Of (E_Queue).E_Time = V_Time loop
         --  remove next event from queue
         E := Front_Of (E_Queue); Pop (E_Queue);

         case E.Kind is

            when Periodic_Request =>
               --  next release time is last release time + period
               P_T_List (E.E_ID).Release_Time := P_T_List (E.E_ID).Release_Time +
                  Periodic_Load_List (E.E_ID).Period;
               --  current deadline is last deadline + period
               P_T_List (E.E_ID).Deadline := P_T_List (E.E_ID).Deadline +
                 Periodic_Load_List (E.E_ID).Period;
               --  remaining execution time starts as the full exec. time
               P_T_List (E.E_ID).Remaining_Exec_Time :=
                 Periodic_Load_List (E.E_ID).Exec_Time;
               --  insert job into dispatching queue
               --  use the complement of its period, to get rate-monotonic order
               Add ((E.E_ID, -Periodic_Load_List (E.E_ID).Period), P_T_Queue);
               --  set up deadline expiration event
               Add ((E.E_ID, Periodic_Deadline, P_T_List (E.E_ID).Deadline), E_Queue);
               --  set up next periodic request event
               Add ((E.E_ID, Periodic_Request, P_T_List (E.E_ID).Release_Time +
                    Periodic_Load_List (E.E_ID).Period), E_Queue);
               Check_Periodic_Request (E.E_ID, Periodic_Load_List (E.E_ID).Exec_Time);

            when Aperiodic_Request =>
               Aperiodic_Task_Requested := Aperiodic_Task_Requested+1;
               ET := Exp_Dist (Mean_Exec_Time);
               if ET=0 then
                  Aperiodic_Task_Completed :=Aperiodic_Task_Completed+1;
               else
                  Add ((E.E_ID, V_Time, ET, ET), A_T_Queue);
               end if;
               IAT := Exp_Dist (Mean_Interarrival_Time);
               Add ((E.E_ID+1, Aperiodic_Request, V_Time+IAT), E_Queue);

            when Periodic_Completion =>
               Delete ((E.E_ID, Periodic_Deadline, P_T_List (E.E_ID).Deadline), E_Queue);
               Pop (P_T_Queue);
               Elapsed_Time := V_Time - Last_Dispatching_Time;
               Total_Periodic := Total_Periodic + Elapsed_Time;
               Check_Periodic_Completion (1);
               Check_Times ("ck4");
               Current_Task := Nul_Task;

            when Aperiodic_Completion =>
               Elapsed_Time := V_Time - Last_Dispatching_Time;
               R_Info.R_Amount := R_Info.R_Amount - Elapsed_Time;
               pragma assert (R_Info.R_Amount >= 0, "aperiodic completion (1)");
               Total_Server := Total_Server + Elapsed_Time;
               Check_Times ("ck5");
               A_T_Info := Front_Of (A_T_Queue); Pop (A_T_Queue);
               A_T_Info.Remaining_Exec_Time :=
                  A_T_Info.Remaining_Exec_Time-Elapsed_Time;
               if A_T_Info.Remaining_Exec_Time>0 then Add (A_T_Info, A_T_Queue);
               else
                  Compute_Response_Time;
               end if;
               Current_Task := Nul_Task;

            when Periodic_Deadline =>
               raise Schedule_Fail;

            when Aperiodic_Server_Replenishment =>
               R_Info.R_Amount := Server_Exec_Time;
               Add ((Server_ID, Aperiodic_Server_Replenishment,
                   V_Time+Server_Period), E_Queue);

         end case;
         if V_Time>=End_of_Test and then Front_Of (E_Queue).Kind=Periodic_Request
         then raise Schedule_Complete; end if;
         if V_Time>=Next_Major_Cycle then
            if CPU_Is_Idle then
               Total_Idle := Total_Idle + (V_Time-Last_Idle_Time);
               Check_Times ("ck6");
               Last_Idle_Time := V_Time;
            end if;
            if Total_Idle > Major_Cycle_Idle and then V_Time > 0 and then
               not Overloaded then
               put (Errs, "PLS: "); Overloaded := true;
               Int_IO.put (Errs, V_Time, 8);
               Flt_IO.put (Errs, Mean_Interarrival_Time, 4, 2, 0);
               Flt_IO.put (Errs, Aperiodic_Load, 4, 2, 0);
               put_line (Errs, " CPU overload ***");
            end if;
            Major_Cycle_Idle := Total_Idle;
            Next_Major_Cycle := Next_Major_Cycle+Major_Period;
         end if;
      end loop;

      Try_Dispatching;
      if Front_Of (E_Queue).E_Time < V_Time then
         raise Program_Error;
      end if;
      V_Time := Front_Of (E_Queue).E_Time;
   end loop;

exception

   when Schedule_Complete =>
      while (not Is_Empty (A_T_Queue)) and then
        Front_Of (A_T_Queue).Release_Time < V_Time-Server_Period loop
         Aperiodic_Task_Missed := Aperiodic_Task_Missed + 1; Pop (A_T_Queue);
      end loop;
      -- compute average response time --
      if Aperiodic_Task_Completed > 0
      then Avg_Resp_Time := Total_Response_Time/float (Aperiodic_Task_Completed);
      else Avg_Resp_Time := float (Erroneous); end if;
      Compute_Statistics;

   when Schedule_Fail =>

      while (not Is_Empty (A_T_Queue)) and then
        Front_Of (A_T_Queue).Release_Time < V_Time - Server_Period loop
         Aperiodic_Task_Missed := Aperiodic_Task_Missed + 1;
         Pop (A_T_Queue);
      end loop;

      Elapsed_Time := V_Time - Last_Dispatching_Time;

      if Current_Task = Nul_Task then
         if CPU_Is_Idle then
            Total_Idle := Total_Idle + Elapsed_Time;
            Check_Times ("ck7");
         end if;
      elsif Current_Task in Periodic_Task_ID_Number then
         Total_Periodic := Total_Periodic + Elapsed_Time;
         Check_Periodic_Completion (2);
         Check_Times ("ck8");
      else
         Total_Server := Total_Server + Elapsed_Time;
         Check_Times ("ck9");
      end if;

      --  display message

      put_line (Errs,
          "PLS: periodic task #" & Periodic_Task_ID_Number'image (E.E_ID)
          & " misses its deadline at time" & integer'image (V_Time));
      put (Errs, "server exec time =");   Int_IO.put (Errs, Server_Exec_Time);
      put (Errs, " server period =");  Int_IO.put (Errs, Server_Period);
      new_line (Errs);
      put (Errs, "remaining exec time ");
      Int_IO.put (Errs, P_T_List(E.E_ID).Remaining_Exec_Time, 8);
      new_line (Errs);
      Dump_Times;
      Avg_Resp_Time := float (Erroneous);
      Compute_Statistics;

      raise;

   when E : others =>
      Put (Errs, "exception " & Exception_Name (E));  New_Line (Errs);
      Put (Errs, Exception_Message (E));  New_Line (Errs);
      put (Errs, "PLS: time: "); put (Errs, V_Time, 7); put (Errs, " e_time: ");
      put (Errs, Front_Of (E_Queue).E_Time, 7); put (Errs, " (");
      put (Errs, Front_Of (E_Queue).Kind);put (Errs, ")"); new_line (Errs);
      Dump_Times;

      raise;

end Polling_Server;

end PLS;
