-- $Id: system_type.adb,v 1.2 2006/11/27 00:14:21 baker Exp baker $

with Ada.Text_IO;
with Ada.Integer_Text_IO;

package body System_Type is

   function ">" (Left, Right: Event) return Boolean is
   begin
      if Left.E_Time > Right.E_Time then
         return True;
      elsif Left.E_Time = Right.E_Time then
         return Left.Kind >= Right.Kind;
      else
         return False;
      end if;
   end ">";

   function ">" (Left, Right: Periodic_Task_Rep) return Boolean is
   begin
      return Left.Priority <= Right.Priority;
   end ">";

   function ">" (Left, Right: Aperiodic_Task_Info) return Boolean is
   begin
      return Left.Release_Time >= Right.Release_Time;
   end ">";

   function ">" (Left, Right: Replenishment_Info) return Boolean is
   begin
      return Left.R_Time >= Right.R_Time;
   end ">";

   procedure Calculate_Major_Period(Periodic_Load_List: in Process_List) is
      P_T_Periods: Positive_Vector(positive(Periodic_Task_ID_Number'first)..
                                     positive(Periodic_Task_ID_Number'last));
      --  ??? bad coding and/or naming style
      --  not only caculates the hyperperiod (major period), but
      --  also initializes several variables that are used to
      --  gather statistics
   begin
      for I in Periodic_Task_ID_Number loop
         P_T_Periods(positive(I)):= Periodic_Load_List(I).Period;
      end loop;
      Major_Period:= LCM (P_T_Periods);
      Aperiodic_Task_Requested:= 0;
      Aperiodic_Task_Completed:= 0;
      Aperiodic_Task_Missed:= 0;
      Total_Response_Time:= 0.0;
      Total_Sq_Resp_Time:= 0.0;
      Maximum_Response_Time:= 0;
      Total_Server:= 0;
   end Calculate_Major_Period;

   --  The following mechanims are for internal consistency
   --  checking. They were added to track down bugs in the simulations.
   --  It is important to keep them active, or risk regressions.

   Ck_Queue: Replenishment_Queue.Queue_Type;
   -- Ck_Queue contains a record of recent server replenishments

   --  The following are used to keep track of what has happened
   --  recently, i.e., since the most recent release time of the
   --  lowest-priority periodic task, which we call the
   --  "current longest period".

   Periodic_Demand,    -- periodic demand that can preempt lowest priority task
   Periodic_Backlog  -- uncompleted portion of the above
     : Integer;

   -- Total_Server - OTS is the total server execution time
   -- in the current longest period, and similarly for
   -- Total_Periodic - OTP, Total_Idle - OTI.

   procedure Init_Times is
   begin
     V_Time := 0;
     Total_Periodic := 0;
     Total_Server := 0;
     Total_Idle := 0;
     Periodic_Demand := 0;
     Periodic_Backlog := 0;
     Elapsed_Time := 0;
     Last_Dispatching_Time := 0;
     Current_Task := Nul_Task;
     Replenishment_Queue.Clear (Ck_Queue);
   end Init_Times;

   procedure Dump_Times is
      --  prints out values of variable for diagnosis of
      --  assertion failures
      use Ada.Text_IO; use Ada.Integer_Text_IO;
   begin
      put (Errs, "V_Time         = "); put (Errs, V_Time, 7);
      new_line (Errs);
      put (Errs, "Total_Idle     = "); put (Errs, Total_Idle, 7);
      new_line (Errs);
      put (Errs, "Total_Periodic (completed) = "); put (Errs, Total_Periodic, 7);
      new_line (Errs);
      put (Errs, "Total_Server (completed) = "); put (Errs, Total_Server, 7);
      new_line (Errs);
      put (Errs, "Periodic_Demand so far ="); put (Errs, Periodic_Demand, 8);
      new_line (Errs);
      put (Errs, "Periodic_Backlog ="); put (Errs, Periodic_Backlog, 8);
      new_line (Errs);
      Dump_Ck;
      Prune_Ck (1300);
   end Dump_Times;

   procedure Check_Times (Msg : String) is
      --  checks several assertions about time values
      D : Integer := V_Time - Total_Idle - Total_Periodic - Total_Server;
      use Ada.Text_IO; use Ada.Integer_Text_IO;
   begin
      --  total elapsed virtual time should exactly equal
      --  the sum of system idle time, total periodic task execution time,
      --  and total aperiodic server execution time
      if (D /= 0) then
         put_line (Errs, Msg & ": time discrepancy of " & Integer'Image (D));
         Dump_Times;
         raise Program_Error;
      end if;
   end Check_Times;

   procedure Check_Periodic_Request
     (New_Task : Task_ID_Number; Exec_Time : Integer) is
   begin
      pragma Assert (Exec_Time > 0, "Check_Periodic_Request");
      if New_Task <= Periodic_Task_ID_Number'Last then
         --  add new request to totals of periodic demand
         --  that can preempt lowest-priority periodic task
         Periodic_Backlog := Periodic_Backlog + Exec_Time;
         Periodic_Demand := Periodic_Demand + Exec_Time;
      end if;
   end Check_Periodic_Request;

   procedure Check_Before_Task_Switch is
      --  reduce pending periodic demand by amount of work completed
      --  by current (assumed periodic) task,
      --  which is about to be preempted
   begin
      if Current_Task > Nul_Task and then
        Current_Task <= Periodic_Task_ID_Number'Last then
         Periodic_Backlog := Periodic_Backlog - Elapsed_Time;
         if Periodic_Backlog < 0 then
            Put_Line (Errs, "**** Check_Before_Task_Switch");
            Put_Line (Errs, "current task = " & Task_Id_Number'Image (Current_Task));
            Dump_Times;
            raise Program_Error;
         end if;
      end if;
   end Check_Before_Task_Switch;


   procedure Check_Periodic_Completion (Which : Integer) is
      --  reduce pending periodic demand by amount of work completed
      --  by current (assumed periodic) task,
      --  which is about to complete
   begin
      pragma Assert (Current_Task in Periodic_Task_ID_Number);
      Periodic_Backlog := Periodic_Backlog - Elapsed_Time;
      if Periodic_Backlog < 0 then
         Put_Line (Errs, "**** Check_Periodic_Completion" & Integer'Image (Which));
         Put_Line (Errs, "current task = " & Task_Id_Number'Image (Current_Task));
         Dump_Times;
         raise Program_Error;
      end if;
   end Check_Periodic_Completion;

   procedure Prune_Ck (I: Integer) is
      -- delete Ck_Queue entries older than I time units from now
      use Replenishment_Queue;
   begin
     while not Is_Empty (Ck_Queue) and then
          Front_Of (Ck_Queue).R_Time < V_Time-I loop
       pop (Ck_Queue);
     end loop;
   end Prune_Ck;

   procedure Dump_Ck is
      -- print out Ck_Queue entries, and total value.
      -- side effect: empties content of Ck_Queue
     R_Temp: Replenishment_Info;
     S: Integer;
     use Replenishment_Queue, Ada.Integer_Text_IO;
   begin
     S := 0;
     put_line (Errs, "recent server replenishments:");
     while not Is_Empty (Ck_Queue) loop
       R_Temp := Front_Of (Ck_Queue); pop (Ck_Queue);
       S := S+R_Temp.R_Amount;
       put (Errs, R_Temp.R_Time, 8); put (Errs, R_Temp.R_Amount, 8); new_line (Errs);
     end loop;
     put (Errs, "total  ="); put (Errs, S, 8); new_line (Errs);
   end Dump_Ck;

   procedure Replenishment_Ck (Server_Exec_Time : Integer) is
      -- add R_Info the the replenishment check queue
     use Replenishment_Queue;
   begin
     add ((R_Info.R_Time, Server_Exec_Time-R_Info.R_Amount), Ck_Queue);
   end Replenishment_Ck;

   --  This is the default type of aperiodic scheduler.
   --  It should be overriddent.

   scheduler : String := "unk";

end System_Type;
