-- %W% %G%

with Text_IO; use Text_IO;
with Cycles; use Cycles;
with Generic_Queue;

package System_Type is

  Number_of_Processes: constant:= 10;

  type Task_ID_Number is new Natural;
  subtype Periodic_Task_ID_Number is Task_ID_Number
        range 1..Number_of_Processes;
  subtype Aperiodic_Task_ID_Number is Task_ID_Number
        range Number_of_Processes+1..Task_ID_Number'last;
  Nul_Task: constant Task_ID_Number:= 0;

  type Process is
    record
      Release_Time: integer;
      Exec_Time   : integer;
      Period      : integer;
    end record;

  type Process_List is array (Periodic_TasK_ID_Number) of Process;

  type Periodic_Task_Rep is
    record
      T_ID    : Periodic_Task_ID_Number;
      Priority: integer;
    end record;

  type Periodic_Task_Info is
    record
      Release_Time       : integer;
      Exec_Time          : integer;
      Deadline           : integer;
      Remaining_Exec_Time: integer;
    end record;

  type Periodic_Task_List is array (Periodic_Task_ID_Number)
        of Periodic_Task_Info;

  type Aperiodic_Task_Info is
    record
      T_ID               : Aperiodic_Task_ID_Number;
      Release_Time       : integer;
      Exec_Time          : integer;
      Remaining_Exec_Time: integer;
    end record;

  type Event_Type is (Aperiodic_Server_Replenishment,
                      Aperiodic_Completion,
                      Periodic_Completion,
                      Periodic_Deadline,
                      Aperiodic_Request,
                      Periodic_Request);
  type Event is
    record
      E_ID  : Task_ID_Number;
      Kind  : Event_Type;
      E_Time: integer;
    end record;

  Schedule_Complete, Schedule_Fail, Program_Error: exception;

  type Replenishment_Info is
     record
        R_Time: integer;
        R_Amount: integer;
     end record;

   package Flt_IO is new Float_IO(float);
   package Int_IO is new Integer_IO(integer);
   package Event_IO is new Enumeration_IO(Event_Type);

   -------------------------------
   -- Event queue instantiation --
   -------------------------------
   function ">" (Left, Right: Event) return Boolean;
   package Event_Queue is new Generic_Queue(Event);

   ---------------------------------------
   -- Periodic task queue instantiation --
   ---------------------------------------
   function ">" (Left, Right: Periodic_Task_Rep) return Boolean;
   package Periodic_Task_Queue is new Generic_Queue (Periodic_Task_Rep);

   ----------------------------------------
   -- Aperiodic task queue instantiation --
   ----------------------------------------
   function ">" (Left, Right: Aperiodic_Task_Info) return Boolean;
   package Aperiodic_Task_Queue is new Generic_Queue (Aperiodic_Task_Info);

   ---------------------------------------
   -- replenishment queue instantiation --
   ---------------------------------------
   function ">" (Left, Right: Replenishment_Info) return Boolean;
   package Replenishment_Queue is new Generic_Queue (Replenishment_Info);

   E_Queue : Event_Queue.Queue_Type;
   P_T_Queue : Periodic_Task_Queue.Queue_Type;
   A_T_Queue : Aperiodic_Task_Queue.Queue_Type;
   R_Queue : Replenishment_Queue.Queue_Type;

   Major_Period, End_of_Test,
   Response_Time, Maximum_Response_Time,
   Aperiodic_Task_Requested, Aperiodic_Task_Completed,
   Aperiodic_Task_Missed                            : Integer;
   Total_Response_Time, Total_Sq_Resp_Time          : Float;

   Current_Task : Task_ID_Number;
   Elapsed_Time,
   Last_Dispatching_Time : Integer;

   --  The following are used to check that we are keeping
   --  accurate track of how time is divided between the periodic
   --  tasks, the aperiodic server, and idle time.

   V_Time,           --  virtual time passed so far
   Total_Periodic,   --  total time spent executing periodic tasks
   Total_Idle,       --  total idle time
   Total_Server      --  total time spent by the aperiodic server
   : Integer;

   procedure Check_Times (Msg : String);
   --  checks time invariant,
   --  that total cpu time consumed
   --  equals total virtual time passed in simulation
   procedure Dump_Times;
   --  for debugging, prints out time values used in checks
   procedure Init_Times;
   --  initializes time values used in checks
   procedure Check_Before_Task_Switch;
   --  call to update check values upon preemption of periodic job
   procedure Check_Periodic_Request
     (New_Task : Task_ID_Number; Exec_Time : Integer);
   --  call to update check values upon arrival of periodic request
   procedure Check_Periodic_Completion (Which : Integer);
   --  call to update check values upon completion of periodic request

   Erroneous : constant Integer := -99999;
   -- a value used to indicate unreliable or erroneous data

   P_T_List : Periodic_Task_List;

   procedure Calculate_Major_Period (Periodic_Load_List: in Process_List);

   Errs: File_Type;

   R_Info : Replenishment_Info;

   -- the following are used to providing debugging information on recent replenishments

   procedure Prune_Ck (I: Integer);
   -- delete Ck_Queue entries older than I time units from now
   procedure Replenishment_Ck (Server_Exec_Time : Integer);
   procedure Dump_Ck;
   -- print out Ck_Queue entries, and total
   -- empties content of Ck_Queue


end System_Type;
